import requests
import time
import threading
from .codecarbon.emissions_tracker import EmissionsTracker
import json
import socket
from .hardware import CPU, GPU, RAM, AppleSiliconChip
from collections import Counter
from .cpu import *
from .gpu import *
from .powermetrics import *
import os

class MacEmmissionTracker():
    """
    A class for tracking energy consumption of hardware devices.

    Attributes:
        energy (float): The total energy consumption.
        energy_distr (dict): A dictionary representing the distribution of energy consumption.
        hardware (list): A list of hardware devices being tracked.
        last_duration (int): The duration of the last measurement.
        output_dir (str): The directory to store the measurement results.

    Methods:
        flush(): Measures the power and energy consumption of the tracked hardware devices.
    """
    def __init__(self, last_duration=1, dashboard=False):
        """
            Initializes a CarbonTracker object.

            Args:
                last_duration (int): The duration of the last recorded energy consumption in hours. Defaults to 1.
        """
        
        self.dash = dashboard
        self.energy = 0
        self.energy_distr = {}
        self.hardware = []
        self.last_duration = last_duration
        self.output_dir = "output_dir"
        #self.carbon_intensity = get_carbon_intensity()
        self.carbon_intensity = 0.0005

        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

        if is_gpu_details_available():
            gpu_devices = GPU.from_utils(GPU._get_gpu_ids)
            self.hardware.append(gpu_devices)
            gpu_names = [n["name"] for n in gpu_devices.devices.get_gpu_static_info()]
            gpu_names_dict = Counter(gpu_names)


        if is_powergadget_available():
            hardware = CPU.from_utils(self.output_dir, "intel_power_gadget")
            self.hardware.append(hardware)
        elif is_rapl_available():
            hardware = CPU.from_utils(self.output_dir, "intel_rapl")
            self.hardware.append(hardware)
        elif (
            is_powermetrics_available()
        ):
            print("Using powermetrics for CPU power tracking.")
        hardware_cpu = AppleSiliconChip.from_utils(
            self.output_dir, chip_part="CPU"
        )
        self.hardware.append(hardware_cpu)

        hardware_gpu = AppleSiliconChip.from_utils(
            self.output_dir, chip_part="GPU"
        )
        self.hardware.append(hardware_gpu)

        print(self.hardware)

        for hw in self.hardware:
            hw.start()
        
    def send_data(self):
        """
        Send tracking data to the server.
        """
        url = "http://127.0.0.1:5001/send_info"
        response = requests.post(url, json=self.data)
        
        try:
            if response.json()["task"] != None:
                self.task = response.json()["task"]
                self.dev_ids = response.json()["dev_ids"]
        except KeyError:
            pass

        # Check response status
        if response.status_code == 200:
            print("Data sent successfully!")
        else:
            print("Failed to send data.")

    def flush(self):
        """
            Flushes the data and calculates the CO2 emissions for each hardware component.

            Returns:
                dict: A dictionary containing the CO2 emissions for each hardware component.
        """
        result = {}
        co2_emissions = {}
        for hw in self.hardware:
            res = hw.measure_power_and_energy(self.last_duration)
            if "cpu" in str(hw).lower():
                result[str(hw)] = res 
            elif "gpu" in str(hw).lower():
                result[str(hw)] = res
        
        print(result)
        for key, value in result.items():
            #print(result[key][1].kWh)
            co2_emissions[key] = value[1].kWh * self.carbon_intensity
        
        if self.dash:
            self.data = {
                "timestamp" : time.time(),
                "energy_consumed": (result["cpu"][1].kWh, result["gpu"][1].kWh),
                "host": 'test_host',
                "task": 'test_task',
                "model": 'test_model',
                "time": self.time,
                "hardware_distr": self.energy_distr,
                "dev_ids": self.dev_ids,
                "start": self.start,
                "mode": self.mode
            }

        return co2_emissions





class EmissionTracker:
    """
    A helper class for tracking emissions and sending data to a server.

    Attributes:
        data (dict): A dictionary to store tracking data.
        dev_ids (list): A list of device IDs.
        power_usg (int): The power usage.
        nvml (bool): Flag indicating if NVML is enabled.
        gpu (bool): Flag indicating if GPU is enabled.
        tracker (EmissionsTracker): An instance of the EmissionsTracker class.
        tracking_info (int): Tracking information.
        running (bool): Flag indicating if tracking is running.
        task (str): The task being tracked.
        hostname (str): The hostname of the client.
        mode (str): The mode of tracking.
        start (bool): Flag indicating if tracking has started.
        time (int): The elapsed time.
        time_start (int): The start time of tracking.

    Methods:
        start_tracking: Start tracking emissions.
        stop_tracking: Stop tracking emissions.
        flushing_thread: Thread for flushing tracking data.
        send_data: Send tracking data to the server.
        flush: Flush tracking data.
    """

    def __init__(self, task="track_new", GPU=False, dashboard=True):
        """
        Initializes a new instance of the ClientHelper class.

        Args:
            task (str, optional): The task to be tracked. Defaults to "track_new".
            GPU (bool, optional): Flag indicating if GPU is enabled. Defaults to False.
        """
        self.data = {}
        self.dev_ids = []
        self.power_usg = 0
        self.nvml = False
        self.gpu = GPU
        self.tracker = EmissionsTracker(save_to_file=False)
        self.tracking_info = 0
        self.running = True
        self.task = task
        self.hostname = socket.gethostname()
        self.mode = "training"
        self.start = True
        self.time = 0
        self.time_start = 0
        self.total_energy_cpu = 0
        self.total_energy_gpu  = 0
        self.dashboard = dashboard

        if not self.dashboard:
            self.carbon_intensity = get_carbon_intensity()

    def start_tracking(self, task, model, mode="training"):
        """
        Start tracking emissions.

        Args:
            task (str): The task to be tracked.
            model: The model being used.
            mode (str, optional): The mode of tracking. Defaults to "training".

        Raises:
            ValueError: If an invalid mode is provided.
        """
        valid_values = {'training', 'inference'}
        
        if mode.lower() not in valid_values:
            raise ValueError('Invalid input. The value must be either "training" or "inference".')
        
        self.task = task
        self.model = model
        self.mode = mode
        self.start = True
        hardware = self.tracker.get_hardware()
        hardware = [str(h) for h in hardware]


        self.data = {"task": self.task, "model": self.model, "mode": self.mode, "hardware": hardware, 'host': self.hostname}

        if self.dashboard:
            self.send_data()

        self.tracker.start()

        self.time_start = time.time()

        
        """if self.dashboard:
            self.flushing = threading.Thread(target=self.flushing_thread, daemon=True)
            self.flushing.start()"""


    def stop_tracking(self):
        """
        Stop tracking emissions.
        """
        self.tracker.stop()
        self.running = False

        if self.flushing is not None:
            self.flushing.join()


    def flushing_thread(self):
        """
        Thread for flushing tracking data.
        """
        if self.mode != "inference":
            while self.running:
                print("Flushing...")
                self.flush()
        
    def send_data(self):
        """
        Send tracking data to the server.
        """
        url = "http://127.0.0.1:5001/send_info"
        response = requests.post(url, json=self.data)
        
        try:
            if response.json()["task"] != None:
                self.task = response.json()["task"]
                self.dev_ids = response.json()["dev_ids"]
        except KeyError:
            pass

        # Check response status
        if response.status_code == 200:
            print("Data sent successfully!")
        else:
            print("Failed to send data.")
    
    def flush(self):
        """
        Flush tracking data.
        
        Returns:
            int: The energy consumed during the flush.
        """

        
        self.tracking_info, self.energy_distr, self.total_energy_cpu, self.total_energy_gpu = self.tracker.flush()
        self.tracking_info = json.loads(self.tracking_info.toJSON())
        self.time += time.time() - self.time_start
        self.time_start = time.time()
        
        print(self.total_energy_cpu, self.total_energy_gpu)
        for key, value in self.energy_distr.items():
            if "cpu" in str(key).lower():
                self.energy_distr[key] = self.total_energy_cpu.kWh
            elif "gpu" in str(key).lower():
                self.energy_distr[key] = self.total_energy_gpu.kWh
        
        print("NEW", self.energy_distr)

        self.data = {
            "timestamp" : self.tracking_info["timestamp"],
            "energy_consumed": self.tracking_info["energy_consumed"],
            "host": self.hostname,
            "task": self.task,
            "model": self.model,
            "time": self.time,
            "hardware_distr": self.energy_distr,
            "dev_ids": self.dev_ids,
            "start": self.start,
            "mode": self.mode
        }


        if self.dashboard:
            self.send_data()
        else:
            return self.tracking_info["energy_consumed"], self.carbon_intensity

        self.start = False

        return self.tracking_info["energy_consumed"]

def get_carbon_intensity(time_dur=None):
    """api_keys = {"electricitymaps": "ESfskPIAkjDJQSPVbKdwJzR1tDM7DHko"}
    elec_map = elect_map.ElectricityMap()      
    elec_map.set_api_key(api_keys["electricitymaps"])  
    carbon_intensity = None
    g_location = geocoder.ip("me")
    carbon_intensity, power_mix = elec_map.carbon_intensity(g_location, time_dur)
    carbon_intensity.address = g_location.address
    ci = carbon_intensity.carbon_intensity
    address = carbon_intensity.address
    return ci, address, power_mix"""

    from entsoe import EntsoePandasClient
    import pandas as pd

    client = EntsoePandasClient(api_key="6d6f58a3-1d96-4ea4-a12d-5816d5e81d6f")

    start = pd.Timestamp.now(tz='Europe/Brussels') - pd.Timedelta(hours=3)
    end = pd.Timestamp.now(tz='Europe/Brussels')

    country_code = 'DE'  # Deutschland

    loads = client.query_generation(country_code, start=start, end=end)

    cis = {"Biomass": 230, "Fossil Brown coal/Lignite": 1167, "Fossil Coal-derived gas":234, "Fossil Gas":572, "Fossil Hard coal":1167, "Fossil Oil":1170, 
           "Geothermal":38, "Hydro Pumped Storage":419, "Hydro Run-of-river and poundage":11, "Hydro Water Reservoir":11,  
            "Other":700, "Other renewable":50, "Solar":35, "Waste": 580, "Wind Offshore": 13, "Wind Onshore": 13}

    carbon_intensity = 0 
    sum_emissions = 0   
    
    power_mix = {}
    

    for i, row in loads.iterrows():
        for index, value in row.items():
            if str(index[0]) in str(list(power_mix.keys())):
                power_mix[index[0]] += value
            else:
                power_mix[index[0]] = value

            sum_emissions += value
            carbon_intensity += cis[index[0]] * value
        break
    
    carbon_intensity = carbon_intensity / sum_emissions

    power_mix = {
        "Biomass": power_mix["Biomass"],
        "Coal": power_mix["Fossil Brown coal/Lignite"] + power_mix["Fossil Hard coal"],
        "Gas": power_mix["Fossil Gas"],
        "Geo": power_mix["Geothermal"],
        "Hydro": power_mix["Hydro Pumped Storage"] + power_mix["Hydro Run-of-river and poundage"] + power_mix["Hydro Water Reservoir"],
        "Oil": power_mix["Fossil Oil"],
        "Other": power_mix["Other"],
        "Solar": power_mix["Solar"],
        "Waste": power_mix["Waste"],
        "Wind": power_mix["Wind Offshore"] + power_mix["Wind Onshore"]
    }
    
    return carbon_intensity/1000, "address", power_mix


import os

from prometheus_client import push_to_gateway
from prometheus_client.exposition import basic_auth_handler

from codecarbon.prometheus.metric_definitions import (
    cpu_energy_gauge,
    cpu_power_gauge,
    duration_gauge,
    emissions_gauge,
    emissions_rate_gauge,
    energy_consumed_gauge,
    gpu_energy_gauge,
    gpu_power_gauge,
    ram_energy_gauge,
    ram_power_gauge,
    registry,
)


class Prometheus:
    def __init__(self, prometheus_url):
        self.prometheus_url = prometheus_url

    def _auth_handler(self, url, method, timeout, headers, data):
        username = os.getenv("PROMETHEUS_USERNAME")
        password = os.getenv("PROMETHEUS_PASSWORD")
        return basic_auth_handler(
            url, method, timeout, headers, data, username, password
        )

    def add_emission(self, carbon_emission: dict):
        """
        Send emissions data to push gateway
        """

        # Save the values of the metrics to the local registry
        duration_gauge.labels(carbon_emission["project_name"]).set(
            int(carbon_emission["duration"])
        )
        emissions_gauge.labels(carbon_emission["project_name"]).set(
            carbon_emission["emissions"]
        )
        emissions_rate_gauge.labels(carbon_emission["project_name"]).set(
            carbon_emission["emissions_rate"]
        )
        cpu_power_gauge.labels(carbon_emission["project_name"]).set(
            carbon_emission["cpu_power"]
        )
        gpu_power_gauge.labels(carbon_emission["project_name"]).set(
            carbon_emission["gpu_power"]
        )
        ram_power_gauge.labels(carbon_emission["project_name"]).set(
            carbon_emission["ram_power"]
        )
        cpu_energy_gauge.labels(carbon_emission["project_name"]).set(
            carbon_emission["cpu_energy"]
        )
        gpu_energy_gauge.labels(carbon_emission["project_name"]).set(
            carbon_emission["gpu_energy"]
        )
        ram_energy_gauge.labels(carbon_emission["project_name"]).set(
            carbon_emission["ram_energy"]
        )
        energy_consumed_gauge.labels(carbon_emission["project_name"]).set(
            carbon_emission["energy_consumed"]
        )

        # Send the new metric values
        push_to_gateway(
            self.prometheus_url,
            job="codecarbon",
            registry=registry,
            handler=self._auth_handler,
        )



"""
The Carbon Tracker module. The following objects/decorators belong to the Public API
"""

from ._version import __version__  # noqa
from .emissions_tracker import (
    EmissionsTracker,
    OfflineEmissionsTracker,
    track_emissions,
)

__all__ = ["EmissionsTracker", "OfflineEmissionsTracker", "track_emissions"]



from time import time

from codecarbon.external.hardware import CPU, GPU, RAM, AppleSiliconChip
from codecarbon.external.logger import logger


class MeasurePowerEnergy:
    """
    Measure power and energy consumption of a hardware component.
    """

    _last_measured_time: int = 0
    _hardware: list
    _pue: float
    _total_cpu_energy: float
    _total_gpu_energy: float
    _total_ram_energy: float
    _total_energy: float
    _cpu_power: float
    _gpu_power: float
    _ram_power: float

    def __init__(self, hardware, pue):
        """
        :param hardware: list of hardware components to measure
        :param pue: Power Usage Effectiveness of the datacenter
        """
        self._last_measured_time = time()
        self._hardware = hardware
        self._pue = pue
        # TODO: Read initial energy values from hardware
        self._total_cpu_energy = 0
        self._total_gpu_energy = 0
        self._total_ram_energy = 0
        self._total_energy = 0
        # Power cant't be read at init because we need time, so we set it to 0
        self._cpu_power = 0
        self._gpu_power = 0
        self._ram_power = 0

    def do_measure(self) -> None:
        for hardware in self._hardware:
            h_time = time.time()
            # Compute last_duration again for more accuracy
            last_duration = time.time() - self._last_measured_time
            power, energy = hardware.measure_power_and_energy(
                last_duration=last_duration
            )
            # Apply the PUE of the datacenter to the consumed energy
            energy *= self._pue
            self._total_energy += energy
            if isinstance(hardware, CPU):
                self._total_cpu_energy += energy
                self._cpu_power = power
                logger.info(
                    f"Energy consumed for all CPUs : {self._total_cpu_energy.kWh:.6f} kWh"
                    + f". Total CPU Power : {self._cpu_power.W} W"
                )
            elif isinstance(hardware, GPU):
                self._total_gpu_energy += energy
                self._gpu_power = power
                logger.info(
                    f"Energy consumed for all GPUs : {self._total_gpu_energy.kWh:.6f} kWh"
                    + f". Total GPU Power : {self._gpu_power.W} W"
                )
            elif isinstance(hardware, RAM):
                self._total_ram_energy += energy
                self._ram_power = power
                logger.info(
                    f"Energy consumed for RAM : {self._total_ram_energy.kWh:.6f} kWh."
                    + f"RAM Power : {self._ram_power.W} W"
                )
            elif isinstance(hardware, AppleSiliconChip):
                if hardware.chip_part == "CPU":
                    self._total_cpu_energy += energy
                    self._cpu_power = power
                    logger.info(
                        f"Energy consumed for AppleSilicon CPU : {self._total_cpu_energy.kWh:.6f} kWh"
                        + f".Apple Silicon CPU Power : {self._cpu_power.W} W"
                    )
                elif hardware.chip_part == "GPU":
                    self._total_gpu_energy += energy
                    self._gpu_power = power
                    logger.info(
                        f"Energy consumed for AppleSilicon GPU : {self._total_gpu_energy.kWh:.6f} kWh"
                        + f".Apple Silicon GPU Power : {self._gpu_power.W} W"
                    )
            else:
                logger.error(f"Unknown hardware type: {hardware} ({type(hardware)})")
            h_time = time.time() - h_time
            logger.debug(
                f"{hardware.__class__.__name__} : {hardware.total_power().W:,.2f} "
                + f"W during {last_duration:,.2f} s [measurement time: {h_time:,.4f}]"
            )
        logger.info(
            f"{self._total_energy.kWh:.6f} kWh of electricity used since the beginning."
        )



from prometheus_client import CollectorRegistry, Gauge

registry = CollectorRegistry()


# TODO: add labelnames
# timestamp: str
# run_id: str
# python_version: str
# longitude: float
# latitude: float
# on_cloud: str = "N"

# TODO: Set up the possible labels
labelnames = [
    "project_name",
    "country_name",
    "country_iso_code",
    "region",
    "cloud_provider",
    "cloud_region",
    "os",
    "codecarbon_version",
    "cpu_model",
    "cpu_count",
    "gpu_model",
    "gpu_count",
    "tracking_mode",
    "ram_total_size",
]

duration_gauge = Gauge(
    "codecarbon_duration",
    "Duration from last measure (s)",
    ["project_name"],
    registry=registry,
)
emissions_gauge = Gauge(
    "codecarbon_emissions",
    "Emissions as CO₂-equivalents [CO₂eq] (kg)",
    ["project_name"],
    registry=registry,
)
emissions_rate_gauge = Gauge(
    "codecarbon_emissions_rate",
    "Emissions divided per duration (Kg/s)",
    ["project_name"],
    registry=registry,
)
cpu_power_gauge = Gauge(
    "codecarbon_cpu_power",
    "CPU power (W)",
    ["project_name"],
    registry=registry,
)
gpu_power_gauge = Gauge(
    "codecarbon_gpu_power",
    "GPU power (W)",
    ["project_name"],
    registry=registry,
)
ram_power_gauge = Gauge(
    "codecarbon_ram_power",
    "RAM power (W)",
    ["project_name"],
    registry=registry,
)
cpu_energy_gauge = Gauge(
    "codecarbon_cpu_energy",
    "Energy used per CPU (kWh)",
    ["project_name"],
    registry=registry,
)
gpu_energy_gauge = Gauge(
    "codecarbon_gpu_energy",
    "Energy used per GPU (kWh)",
    ["project_name"],
    registry=registry,
)
ram_energy_gauge = Gauge(
    "codecarbon_ram_energy",
    "Energy used per RAM (kWh)",
    ["project_name"],
    registry=registry,
)
energy_consumed_gauge = Gauge(
    "codecarbon_energy_consumed",
    "Sum of cpu_energy, gpu_energy and ram_energy (kW)",
    ["project_name"],
    registry=registry,
)



from flask import Flask, render_template, request, jsonify
#import geocoder
import statistics
from collections import Counter
import time
import threading
import json
from functools import reduce
from .db_app import db, app, train_devices, inference_devices, hostgpus, Task, Model, Hostgpu, Trackinginfo, Inferenceinfo, Powermix


    
### DATABASE FUNCTIONS ###
def add_init_entry(mode, model, task, hardware, host):
    """
    Adds an initial entry to the database for a given mode, model, task, hardware, and host.

    Parameters:
    - mode (str): The mode of the entry (e.g., "inference").
    - model (str): The name of the model.
    - task (str): The name of the task.
    - hardware (list): A list of hardware devices.
    - host (str): The name of the host.

    Returns:
    - updated_task_name (str): The updated task name.
    - dev_ids (dict): A dictionary mapping hardware devices to their corresponding IDs.
    """
    updated_task_name = ""
    existing_model = Model.query.filter_by(model_name=model, task=task).first()
    if not existing_model:
        updated_task_name = task
    else:
        i = 1
        while True:
            updated_task_name = f"{task}_{i}"
            existing_model = Model.query.filter_by(model_name=model, task=updated_task_name).first()
            if not existing_model:
                break
            i += 1

    inference_info = Inferenceinfo.query.filter_by(task_name_inf=task, model_inf=model).first()  
    if updated_task_name != task and mode == "inference" and not inference_info:
        updated_task_name = task
    
    elif inference_info and mode == "inference":
        i = 1
        while True:
            updated_task_name = f"{task}_{i}"
            inference_info = Inferenceinfo.query.filter_by(task_name_inf=updated_task_name, model_inf=model).first()
            if not inference_info:
                break
            i += 1

    existing_model = Model.query.filter_by(model_name=model, task=updated_task_name).first()
    existing_task = Task.query.filter_by(task=updated_task_name).first()

    if not existing_model:
        new_model = Model(model_name=model, task=updated_task_name)
        db.session.add(new_model)
        db.session.commit()

    if not existing_task:
        new_task = Task(task=updated_task_name)
        db.session.add(new_task)
        db.session.commit()

    dev_ids = {}
    for dev in hardware:
        if "gpu" in dev.lower() or "cpu" in dev.lower():
            host_exists = Hostgpu.query.filter_by(host=host, gpu_cpu=dev).first()
            if host_exists is None:
                new_host_gpu = Hostgpu(
                host=host,
                gpu_cpu=dev
            )
                db.session.add(new_host_gpu)
                db.session.commit()

            query = db.session.query(Hostgpu.id).filter(Hostgpu.host == host, Hostgpu.gpu_cpu == dev).first()
            if host_exists and existing_task:
                dev_ids[dev] = query[0]
            else:
                if query:
                    dev_ids[dev] = query[0]
                else:
                    dev_ids[dev] = new_host_gpu.id

                # Associate the task with the host GPU
                if existing_task and query:
                    new_hostgpus = hostgpus.insert().values(task_id=updated_task_name, hostgpu_host=query[0])
                elif not existing_task and not query:
                    new_hostgpus = hostgpus.insert().values(task_id=new_task.task, hostgpu_host=new_host_gpu.id)
                elif not existing_task and query:
                    new_hostgpus = hostgpus.insert().values(task_id=new_task.task, hostgpu_host=query[0])

                db.session.execute(new_hostgpus)
                db.session.commit()

    return updated_task_name, dev_ids

def add_train_entry(timestamp, carb_int=None, power_usg=None, CO2_usg=None, task=None, 
                model=None, elapsed_time=None, dev_ids=None, hardware_distr=None, start=None):
    """
    Add a train entry to the database.

    Args:
        timestamp (datetime): The timestamp of the train entry.
        carb_int (float, optional): The carbon intensity. Defaults to None.
        power_usg (float, optional): The power usage. Defaults to None.
        CO2_usg (float, optional): The CO2 usage. Defaults to None.
        task (str, optional): The task name. Defaults to None.
        model (str, optional): The model name. Defaults to None.
        elapsed_time (float, optional): The elapsed time. Defaults to None.
        dev_ids (dict, optional): The dictionary of device IDs. Defaults to None.
        hardware_distr (dict, optional): The dictionary of hardware distribution. Defaults to None.
        start (bool, optional): Indicates if it's the start of a train. Defaults to None.
    """
    # Create a data entry and add it to the database
    new_entry = Trackinginfo(
        timestamp=timestamp,
        carbon_intensity=carb_int,
        power_usage=power_usg,
        CO2_usage=CO2_usg,
        task_name=task,
        model=model,
        elapsed_time=elapsed_time,
        task_model=f"{task}_{model}"
    )
    db.session.add(new_entry)
    db.session.commit()
    
    for dev, dev_id in dev_ids.items():  
        new_hostdevice = train_devices.insert().values(tracking_info_id=new_entry.id, hostgpu_id=dev_id, CO2_usage=hardware_distr[dev]*carb_int, power_usg=hardware_distr[dev])
        db.session.execute(new_hostdevice)
        db.session.commit()

def add_inference_entry(timestamp, carb_int=None, power_usg=None, CO2_usg=None, task=None,
              model=None, elapsed_time=None, dev_ids=None, hardware_distr=None):
    """
    Add a train entry to the database.

    Args:
        timestamp (datetime): The timestamp of the train entry.
        carb_int (float, optional): The carbon intensity. Defaults to None.
        power_usg (float, optional): The power usage. Defaults to None.
        CO2_usg (float, optional): The CO2 usage. Defaults to None.
        task (str, optional): The task name. Defaults to None.
        model (str, optional): The model name. Defaults to None.
        elapsed_time (float, optional): The elapsed time. Defaults to None.
        dev_ids (dict, optional): The dictionary of device IDs. Defaults to None.
        hardware_distr (dict, optional): The dictionary of hardware distribution. Defaults to None.
        start (bool, optional): Indicates if it's the start of a train. Defaults to None.
    """

    # Create a data entry and add it to the database
    new_entry = Inferenceinfo(
        timestamp_inf=timestamp,
        carbon_intensity_inf=carb_int,
        power_usage_inf=power_usg,
        CO2_usage_inf=CO2_usg,
        task_name_inf=task,
        model_inf=model,
        elapsed_time_inf=elapsed_time,
        task_model_inf=f"{task}_{model}"
    )
    db.session.add(new_entry)
    db.session.commit()

    for dev, dev_id in dev_ids.items():
        new_hostdevice = inference_devices.insert().values(inference_info_id=new_entry.id, hostgpu_id=dev_id, CO2_usage=hardware_distr[dev])
        db.session.execute(new_hostdevice)
        db.session.commit()

def get_power_mix():
    """
    Retrieves the power mix.

    Returns:
    - power_mix (dict): The power mix.
    """
    power_mix = Powermix.query.order_by(Powermix.timestamp.desc()).first()
    power_mix = power_mix.powermix.replace("\'", "\"")
    return json.loads(power_mix)

### HELPER FUNCTIONS ###
def get_initial_data():
    tracking_info_entries = Trackinginfo.query.all()
    task_entries = Task.query.all()
    hostgpu_entries = Hostgpu.query.all()
    model_entries = Model.query.all()   
    #hostgpus_entries = db.session.query(hostgpus)

    return tracking_info_entries, task_entries, hostgpu_entries, model_entries

def get_carbon_data(update_data):
    """
    Retrieves carbon data based on the provided update data.

    Args:
        update_data (dict): The update data containing information about tasks, GPUs, hosts, and models.

    Returns:
        tuple: A tuple containing the following carbon data:
            - tracking_info (dict): A dictionary containing tracking information for each task.
            - carb_int_average (float): The average carbon intensity.
            - power_usg_sum (float): The sum of power usage.
            - CO2_usg_sum (float): The sum of CO2 usage.
            - elapsed_time (dict): A dictionary containing elapsed time for each task.
            - train_info (dict): A dictionary containing training information for each task.
    """
    tracking_info = {}
    carb_int = []
    power_usg = []
    CO2_usg = []
    elapsed_time = {}

    for task in update_data['tasks']:
        curr_tracking_info = db.session.execute(db.select(train_devices.c.CO2_usage, train_devices.c.power_usg, Trackinginfo.carbon_intensity,
                                                          Trackinginfo.model, Trackinginfo.elapsed_time,Trackinginfo.task_model, 
                                                          Hostgpu.gpu_cpu, Hostgpu.host, train_devices.c.tracking_info_id) \
                                                         .join(train_devices, Trackinginfo.id == train_devices.c.tracking_info_id)\
                                                         .join(Hostgpu, Hostgpu.id == train_devices.c.hostgpu_id) \
                .filter(Trackinginfo.task_name == task) \
                .filter(Hostgpu.gpu_cpu.in_(update_data["gpus"]))\
                .filter(Hostgpu.host.in_(update_data["hosts"]))\
                .filter(Trackinginfo.task_model.in_(update_data["models"]))\
                ).fetchall()

        tracking_info[task] = curr_tracking_info

    train_info = {key: [(entry[0], entry[7], entry[6]) for entry in value] for key, value in tracking_info.items()}
    aggregated_data = {}

    for key, values in tracking_info.items():
        visited = set()
        aggregated_data[key] = []
        for value in values:
            group_key = value[-1]  
            len_entry = len(values) / len(values[values[-1]==group_key])
            if group_key not in visited:
                    aggregated_data[key].append([
                        value[0], value[1], value[2], value[3], value[4], value[5], value[6]
                    ])
                    visited.add(group_key)
            else:
                aggregated_data[key][-1][:2] = [sum(val) for val in zip(aggregated_data[key][-1][:2], value[:2])]
                aggregated_data[key][-1][2] = value[2]
                aggregated_data[key][-1][3] = value[3]
                aggregated_data[key][-1][4] += value[4]
                aggregated_data[key][-1][5:7] = value[5:7]

        aggregated_data = {key: [[val[i] / len_entry if i == 4 else val[i] 
                                  for i in range(len(val))] for val in values] for key, values in aggregated_data.items()}

    for key, value in aggregated_data.items():
        elapsed_time[key] = [(val[4], val[3]) for val in value]
        tracking_info[key] = [(val[0], val[3], val[5]) for val in value]
        carb_int.append([val[2] for val in value])
        power_usg.append([val[1] for val in value])
        CO2_usg.append([val[0] for val in value])
    
    # Calculate the aggregations of key values
    CO2_usg = reduce(lambda x,y: x+y, CO2_usg)
    power_usg = reduce(lambda x,y: x+y, power_usg)
    carb_int = reduce(lambda x,y: x+y, carb_int)

    print(carb_int)
    try:
        carb_int_average = statistics.mean(carb_int)
    except statistics.StatisticsError:
        carb_int_average=0
        print("No train data yet, skipping, ....")


    try:
        sum_co2 = CO2_usg[-1]
    except IndexError:
        return tracking_info, carb_int_average, 0, 0, elapsed_time, train_info

    sum_power = power_usg[-1]  
    for i in range(len(CO2_usg)):
        if i==0:
            last = CO2_usg[i] 
            last_p = power_usg[i]
            continue
        else:
            if CO2_usg[i] < last:
                sum_co2 += last
                sum_power += last_p
        last = CO2_usg[i]
        last_p = power_usg[i]

    return tracking_info, carb_int_average, sum_power, sum_co2, elapsed_time, train_info

def get_inf_data(update_data):
    """
    Retrieves inference information based on the provided update_data.

    Args:
        update_data (dict): A dictionary containing the update data.

    Returns:
        tuple: A tuple containing three elements:
            - inference_info_averages (list): A list of tuples containing the average CO2 usage for every unique task and model combination.
            - elapsed_time_inf (dict): A dictionary containing the elapsed time for each task.
            - inference_info (dict): A dictionary containing the inference information for each task.

    """
    inference_info = {}
    elapsed_time_inf = {}

    for task in update_data['tasks']:
        curr_inf_info = db.session.execute(db.select(inference_devices.c.CO2_usage, Inferenceinfo.carbon_intensity_inf, Inferenceinfo.power_usage_inf, Inferenceinfo.model_inf, 
                                                     Inferenceinfo.elapsed_time_inf, Inferenceinfo.task_model_inf, Hostgpu.gpu_cpu)\
                                                    .join(inference_devices, Inferenceinfo.id == inference_devices.c.inference_info_id)\
                                                    .join(Hostgpu, Hostgpu.id == inference_devices.c.hostgpu_id) \
            .filter(Inferenceinfo.task_name_inf == task) \
            .filter(Hostgpu.gpu_cpu.in_(update_data["gpus"]))\
            .filter(Hostgpu.host.in_(update_data["hosts"]))\
            .filter(Inferenceinfo.task_model_inf.in_(update_data["models"]))\
            ).fetchall()
        inference_info[task] = curr_inf_info

    # Calculate the average CO2_usage for every unique task and model combination
    inference_info_averages = []
    for task, info in inference_info.items():
        model_co2_usage = {}
        for entry in info:
            model = entry[3]
            co2_usage = entry[0]
            if model not in model_co2_usage:
                model_co2_usage[model] = [co2_usage]
            else:
                model_co2_usage[model].append(co2_usage)
        for model, co2_usage_list in model_co2_usage.items():
            average_co2_usage = sum(co2_usage_list) / len(co2_usage_list)
            inference_info_averages.append((task, model, average_co2_usage))

    inference_info = {key: [(entry[0], entry[-1]) for entry in value] for key, value in inference_info.items()}

    return inference_info_averages, elapsed_time_inf, inference_info
    
def get_carbon_intensity(time_dur=None):
    """api_keys = {"electricitymaps": "ESfskPIAkjDJQSPVbKdwJzR1tDM7DHko"}
    elec_map = elect_map.ElectricityMap()      
    elec_map.set_api_key(api_keys["electricitymaps"])  
    carbon_intensity = None
    g_location = geocoder.ip("me")
    carbon_intensity, power_mix = elec_map.carbon_intensity(g_location, time_dur)
    carbon_intensity.address = g_location.address
    ci = carbon_intensity.carbon_intensity
    address = carbon_intensity.address
    return ci, address, power_mix"""

    from entsoe import EntsoePandasClient
    import pandas as pd

    client = EntsoePandasClient(api_key="6d6f58a3-1d96-4ea4-a12d-5816d5e81d6f")

    start = pd.Timestamp.now(tz='Europe/Brussels') - pd.Timedelta(hours=3)
    end = pd.Timestamp.now(tz='Europe/Brussels')

    country_code = 'DE'  # Deutschland

    loads = client.query_generation(country_code, start=start, end=end)

    cis = {"Biomass": 230, "Fossil Brown coal/Lignite": 1167, "Fossil Coal-derived gas":234, "Fossil Gas":572, "Fossil Hard coal":1167, "Fossil Oil":1170, 
           "Geothermal":38, "Hydro Pumped Storage":419, "Hydro Run-of-river and poundage":11, "Hydro Water Reservoir":11,  
            "Other":700, "Other renewable":50, "Solar":35, "Waste": 580, "Wind Offshore": 13, "Wind Onshore": 13}

    carbon_intensity = 0 
    sum_emissions = 0   
    
    power_mix = {}
    

    for i, row in loads.iterrows():
        for index, value in row.items():
            if str(index[0]) in str(list(power_mix.keys())):
                power_mix[index[0]] += value
            else:
                power_mix[index[0]] = value

            sum_emissions += value
            carbon_intensity += cis[index[0]] * value
        break
    
    carbon_intensity = carbon_intensity / sum_emissions

    power_mix = {
        "Biomass": power_mix["Biomass"],
        "Coal": power_mix["Fossil Brown coal/Lignite"] + power_mix["Fossil Hard coal"],
        "Gas": power_mix["Fossil Gas"],
        "Geo": power_mix["Geothermal"],
        "Hydro": power_mix["Hydro Pumped Storage"] + power_mix["Hydro Run-of-river and poundage"] + power_mix["Hydro Water Reservoir"],
        "Oil": power_mix["Fossil Oil"],
        "Other": power_mix["Other"],
        "Solar": power_mix["Solar"],
        "Waste": power_mix["Waste"],
        "Wind": power_mix["Wind Offshore"] + power_mix["Wind Onshore"]
    }
    
    return carbon_intensity/1000, "address", power_mix


### FUNCTIONS TO GET DYNAMIC CHART DATA FOR DASHBOARD ###
def get_chart_data(data, largest_task_length):
    chart_data = {
        "type": "line",
        "data": {
            "labels": [],
            "datasets": []
        },
        "options": {
            "maintainAspectRatio": False,
            "legend": {
                "display": True, 
                "labels": {
                    "fontStyle": "normal"
                }
            },
            "title": {
                "fontStyle": "normal"
            },
            "scales": {
                "xAxes": [
                    {
                        "gridLines": {
                            "color": "rgb(234, 236, 244)",
                            "zeroLineColor": "rgb(234, 236, 244)",
                            "drawBorder": False,
                            "drawTicks": False,
                            "borderDash": ["2"],
                            "zeroLineBorderDash": ["2"],
                            "drawOnChartArea": False
                        },
                        "ticks": {
                            "fontColor": "#858796",
                            "fontStyle": "normal",
                            "padding": 20
                        }
                    }
                ],
                "yAxes": [
                    {
                        "gridLines": {
                            "color": "rgb(234, 236, 244)",
                            "zeroLineColor": "rgb(234, 236, 244)",
                            "drawBorder": False,
                            "drawTicks": False,
                            "borderDash": ["2"],
                            "zeroLineBorderDash": ["2"]
                        },
                        "ticks": {
                            "fontColor": "#858796",
                            "fontStyle": "normal",
                            "padding": 20
                        }
                    }
                ]
            }
        }
    }

    colors = [
        "rgba(78, 115, 223, 0.05)",
        "rgba(223, 78, 115, 0.05)",
        "rgba(115, 223, 78, 0.05)",
        "rgba(223, 115, 78, 0.05)",
        "rgba(78, 223, 115, 0.05)",
        "rgba(115, 78, 223, 0.05)",
        "rgba(223, 223, 78, 0.05)",
        "rgba(78, 223, 223, 0.05)",
        "rgba(223, 78, 223, 0.05)",
        "rgba(115, 115, 223, 0.05)",
        "rgba(223, 115, 115, 0.05)",
        "rgba(115, 223, 223, 0.05)",
        "rgba(223, 223, 115, 0.05)",
        "rgba(78, 115, 78, 0.05)",
        "rgba(115, 78, 78, 0.05)",
        "rgba(223, 78, 223, 0.05)",
        "rgba(78, 223, 78, 0.05)",
        "rgba(223, 115, 223, 0.05)",
        "rgba(115, 223, 115, 0.05)",
        "rgba(223, 223, 223, 0.05)"
    ]

    datasets = {}         
    for key in data.keys():
        for i in range(len(data[key])):
            curr_model = data[key][i][1]
            label = key + "_" + data[key][i][1]
            
            # Add all values to the dictionary at entry curr_model
            values = [value[0] for value in data[key] if value[1] == curr_model]
            datasets[label] = values

    for i, (label, dataset) in enumerate(datasets.items()):
        chart_data["data"]["datasets"].append({
            "label": label,
            "fill": True,
            "data": dataset,
            "backgroundColor": colors[i % len(colors)],
            "borderColor": colors[i % len(colors)].replace("0.05", "1"),
            "borderWidth": 3,
        })

    chart_data["data"]["labels"] = [i for i in range(1, largest_task_length + 1)]

    return chart_data

def get_chart_data_inf(inf_data):
    chart_data_inf = {
        "type": "bar",
        "data": {
            "labels": [],
            "datasets": []
        },
        "options": {
            "maintainAspectRatio": False,
            "legend": {
                "display": True, 
                "labels": {
                    "fontStyle": "normal"
                }
            },
            "title": {
                "fontStyle": "normal"
            },
            "scales": {
                "xAxes": [
                    {
                        "gridLines": {
                            "color": "rgb(234, 236, 244)",
                            "zeroLineColor": "rgb(234, 236, 244)",
                            "drawBorder": False,
                            "drawTicks": False,
                            "borderDash": ["2"],
                            "zeroLineBorderDash": ["2"],
                            "drawOnChartArea": False
                        },
                        "ticks": {
                            "fontColor": "#858796",
                            "fontStyle": "normal",
                            "padding": 20
                        }
                    }
                ],
                "yAxes": [
                    {
                        "gridLines": {
                            "color": "rgb(234, 236, 244)",
                            "zeroLineColor": "rgb(234, 236, 244)",
                            "drawBorder": False,
                            "drawTicks": False,
                            "borderDash": ["2"],
                            "zeroLineBorderDash": ["2"]
                        },
                        "ticks": {
                            "fontColor": "#858796",
                            "fontStyle": "normal",
                            "padding": 20,
                            "beginAtZero": True 
                        }
                    }
                ]
            }
        }
    }

    colors = [
        "rgba(78, 115, 223, 0.05)",
        "rgba(223, 78, 115, 0.05)",
        "rgba(115, 223, 78, 0.05)",
        "rgba(223, 115, 78, 0.05)",
        "rgba(78, 223, 115, 0.05)",
        "rgba(115, 78, 223, 0.05)",
        "rgba(223, 223, 78, 0.05)",
        "rgba(78, 223, 223, 0.05)",
        "rgba(223, 78, 223, 0.05)",
        "rgba(115, 115, 223, 0.05)",
        "rgba(223, 115, 115, 0.05)",
        "rgba(115, 223, 223, 0.05)",
        "rgba(223, 223, 115, 0.05)",
        "rgba(78, 115, 78, 0.05)",
        "rgba(115, 78, 78, 0.05)",
        "rgba(223, 78, 223, 0.05)",
        "rgba(78, 223, 78, 0.05)",
        "rgba(223, 115, 223, 0.05)",
        "rgba(115, 223, 115, 0.05)",
        "rgba(223, 223, 223, 0.05)"
    ]

    for i, elem in enumerate(inf_data):
        task, model, data = elem
        chart_data_inf["data"]["datasets"].append({
            "label": task + "_" + model,
            "fill": True,
            "data": [data],
            "backgroundColor": colors[i % len(colors)],
            "borderColor": colors[i % len(colors)].replace("0.05", "1"),
            "borderWidth": 3
        })


    return chart_data_inf

def get_chart_data_power_mix(power_mix):
    chart_data_power_mix = {
        "type": "doughnut",
        "data": {
            "labels": list(power_mix.keys()),
            "datasets": [
                {
                    "data": list(power_mix.values()),
                    "backgroundColor": ["darkgreen", "darkblue", "yellow", "lightblue", "red", "lightgreen", 
                                        "brown", "grey", "black", "white", "purple", "orange", "pink", "cyan", 
                                        "magenta", "lime", "teal", "indigo", "maroon", "navy", "olive", "silver", 
                                        "aqua", "fuchsia", "green"]
                }
            ]
        },
        "options": {
            "responsive": True,
            "maintainAspectRatio": False,
            "legend": {
                "display": False
            }
        }
    }
    
    return chart_data_power_mix

def get_chart_data_gpu_cpu(data, largest_task_length):
    chart_data = {
        "type": "line",
        "data": {
            "labels": [],
            "datasets": []
        },
        "options": {
            "maintainAspectRatio": False,
            "legend": {
                "display": True, 
                "labels": {
                    "fontStyle": "normal"
                }
            },
            "title": {
                "fontStyle": "normal"
            },
            "scales": {
                "xAxes": [
                    {
                        "gridLines": {
                            "color": "rgb(234, 236, 244)",
                            "zeroLineColor": "rgb(234, 236, 244)",
                            "drawBorder": False,
                            "drawTicks": False,
                            "borderDash": ["2"],
                            "zeroLineBorderDash": ["2"],
                            "drawOnChartArea": False
                        },
                        "ticks": {
                            "fontColor": "#858796",
                            "fontStyle": "normal",
                            "padding": 20
                        }
                    }
                ],
                "yAxes": [
                    {
                        "gridLines": {
                            "color": "rgb(234, 236, 244)",
                            "zeroLineColor": "rgb(234, 236, 244)",
                            "drawBorder": False,
                            "drawTicks": False,
                            "borderDash": ["2"],
                            "zeroLineBorderDash": ["2"]
                        },
                        "ticks": {
                            "fontColor": "#858796",
                            "fontStyle": "normal",
                            "padding": 20
                        }
                    }
                ]
            }
        }
    }

    colors = [
        "rgba(78, 115, 223, 0.05)",
        "rgba(223, 78, 115, 0.05)",
        "rgba(115, 223, 78, 0.05)",
        "rgba(223, 115, 78, 0.05)",
        "rgba(78, 223, 115, 0.05)",
        "rgba(115, 78, 223, 0.05)",
        "rgba(223, 223, 78, 0.05)",
        "rgba(78, 223, 223, 0.05)",
        "rgba(223, 78, 223, 0.05)",
        "rgba(115, 115, 223, 0.05)",
        "rgba(223, 115, 115, 0.05)",
        "rgba(115, 223, 223, 0.05)",
        "rgba(223, 223, 115, 0.05)",
        "rgba(78, 115, 78, 0.05)",
        "rgba(115, 78, 78, 0.05)",
        "rgba(223, 78, 223, 0.05)",
        "rgba(78, 223, 78, 0.05)",
        "rgba(223, 115, 223, 0.05)",
        "rgba(115, 223, 115, 0.05)",
        "rgba(223, 223, 223, 0.05)"
    ]

    for i, (label, dataset) in enumerate(data.items()):
        chart_data["data"]["datasets"].append({
            "label": label,
            "fill": True,
            "data": dataset,
            "backgroundColor": colors[i % len(colors)],
            "borderColor": colors[i % len(colors)].replace("0.05", "1"),
            "borderWidth": 3,
        })

    chart_data["data"]["labels"] = [i for i in range(1, largest_task_length + 1)]

    return chart_data

class Dashboard:
    def __init__(self):
        self.app = app
        self.db = db
    ### ROUTES ###
    @app.route('/receive_update', methods=['POST', 'GET'])
    def receive_update():
        """
        Receives update data and generates chart data for carbon tracking.

        Returns:
            JSON: A JSON object containing chart data for carbon tracking, energy consumption, carbon intensity, CO2 emissions,
                power mix, and GPU/CPU information.
        """
        data = request.get_json()
        #print("DATA", data)
        if any(not data_list for data_list in data.values()):
                return jsonify(chart_data_carb=get_chart_data({},0), chart_data_time=get_chart_data({},0), chart_data_inf=get_chart_data_inf([]), energy=0, 
                    carb_int=0, co2=0, chart_data_power_mix=get_chart_data_power_mix(get_power_mix()), chart_data_gpu_cpu=get_chart_data_gpu_cpu({}, 0))   
        
        tracking_info, carb_int, energy, co2, elapsed_time_info, train_info = get_carbon_data(data)
        inference_info_avg, elapsed_time_inf, inference_info  = get_inf_data(data)
        #_, _, power_mix = get_carbon_intensity()
        power_mix = get_power_mix()
        gpu_cpu_info = {}
        # Function to merge values
        def merge_values(d):
            for key, value in d.items():
                for v, h, g in value:
                    curr = f"{h}:{g}"
                    if curr in gpu_cpu_info:
                        gpu_cpu_info[curr].append((v))
                    else:
                        gpu_cpu_info[curr] = [(v)]

        # Apply the function to each dictionary
        merge_values(train_info)

        task_models = [item[2] for sublist in tracking_info.values() for item in sublist]
        try:
            _ ,max_length = Counter(task_models).most_common(1)[0]
        except IndexError:
            max_length = 0

        if max_length != 0:
            chart_data_carb = get_chart_data(tracking_info, max_length)
            chart_data_time = get_chart_data(elapsed_time_info, max_length)
            chart_data_gpu_cpu = get_chart_data_gpu_cpu(gpu_cpu_info, len(gpu_cpu_info[list(gpu_cpu_info.keys())[0]])) 
        else:
            chart_data_carb = get_chart_data({}, 0)
            chart_data_time = get_chart_data({}, 0)
            chart_data_gpu_cpu = get_chart_data_gpu_cpu({}, 0)
            
        chart_data_inf = get_chart_data_inf(inference_info_avg)
        chart_data_power_mix = get_chart_data_power_mix(power_mix)
        

        return jsonify(chart_data_carb=chart_data_carb, chart_data_time=chart_data_time, chart_data_inf=chart_data_inf, energy=energy, 
                    carb_int=carb_int*1000, co2=co2, chart_data_power_mix=chart_data_power_mix, chart_data_gpu_cpu=chart_data_gpu_cpu)
    
    @app.route('/')
    def index():
        """
        Renders the index.html template with the necessary data for the webpage.

        Returns:
            The rendered index.html template with the required data.
        """
        chart_data = get_chart_data({}, 0)
        chart_data_inf = get_chart_data_inf([])
        chart_data_power_mix = get_chart_data_power_mix(get_power_mix())

        
        _ , tasks, hostgpu, models = get_initial_data()
        unique_hosts = list(set(host.host for host in hostgpu))
        unique_tasks = list(set(task.task for task in tasks))

        return render_template('index.html', data={'track_info':{}, 'tasks':tasks, 'hostgpus':hostgpu, 'unique_hosts':unique_hosts, 'unique_tasks':unique_tasks, 'models': models,
                                                'labels':[], 'chart_data':chart_data, 'chart_data_inf':chart_data_inf, 'chart_data_power_mix':chart_data_power_mix})

    @app.route('/send_info', methods=['POST'])
    def receive_info():
        """
        Receives information from a POST request and processes it accordingly.
        
        If the received data contains 'hardware' key, it adds an initialization entry to the tracker
        and returns a JSON response with the task and device IDs.
        
        If the received data contains 'timestamp' key, it adds a training or inference entry to the tracker
        based on the 'mode' value and returns a JSON response indicating the success of data reception.
        
        If the received data does not contain 'hardware' or 'timestamp' key, it returns a JSON response
        indicating that the data was not received.
        """
        data = request.get_json()

        if 'hardware' in data:
            mode = data['mode']
            model = data['model']
            task = data['task']
            hardware = data['hardware']
            host = data['host']

            task, dev_ids = add_init_entry(mode=mode, model=model, task=task, hardware=hardware, host=host)

            return jsonify({"message": "Data received successfully", "task": task, "dev_ids": dev_ids})

        elif 'timestamp' in data:
            timestamp = data['timestamp']
            power_usage = data['energy_consumed']
            host = data['host']
            model = data['model']
            elapsed_time = data['time']
            task = data['task']
            hardware_distr = data['hardware_distr']
            dev_ids = data['dev_ids']
            start = data['start']

        else:
            return jsonify({"message": "Data not received"})
        
        if data["mode"] == "training":
            add_train_entry(carb_int=get_carbon_intensity()[0], power_usg=power_usage, CO2_usg=get_carbon_intensity()[0]*power_usage, 
                    task=task, timestamp=timestamp, elapsed_time=elapsed_time, model=model, dev_ids=dev_ids,hardware_distr=hardware_distr, start=start)

            return jsonify({"message": "Data received successfully"})

        elif data["mode"] == "inference":
            add_inference_entry(carb_int=get_carbon_intensity()[0], power_usg=power_usage, CO2_usg=get_carbon_intensity()[0]*power_usage, 
                        task=task, timestamp=timestamp, elapsed_time=elapsed_time, model=model, dev_ids=dev_ids, hardware_distr=hardware_distr)
            
            return jsonify({"message": "Data received successfully"})
        else:
            return jsonify({"message": "Data not received"})

    @app.route('/delete', methods=['GET'])
    def delete(task, model):
        """
        Deletes a task and model from the database.

        Args:
            task (str): The name of the task to delete.
            model (str): The name of the model to delete.

        Returns:
            str: A string indicating the success of the deletion.
        """
        task = Task.query.filter_by(task=task).first()
        model = Model.query.filter_by(model_name=model).first()
        db.session.delete(task)
        db.session.delete(model)
        db.session.commit()

        return "Task and model deleted successfully"

    def start(self):
        def update_power_mix():
            """
            Updates the power mix.
            """
            while True:
                _, _, power_mix = get_carbon_intensity()

                new_power_mix = Powermix(
                    timestamp=time.time(),
                    powermix=str(power_mix)
                )

                with app.app_context():
                    db.session.add(new_power_mix)
                    db.session.commit()

                time.sleep(3600)

        with app.app_context():
            db.create_all()

            update_power_mix = threading.Thread(target=update_power_mix, daemon=True)
            update_power_mix.start()

        app.run(debug=True, port=5001)






import os
import re
import shutil
import subprocess
import sys
from typing import Dict

import numpy as np

from .util import detect_cpu_model
from .logger import logger


def is_powermetrics_available():
    try:
        ApplePowermetrics()
        response = _has_powermetrics_sudo()
        return response
    except Exception as e:
        logger.debug(
            "Not using PowerMetrics, an exception occurred while instantiating"
            + f" Powermetrics : {e}",
        )
        return False


def _has_powermetrics_sudo():
    process = subprocess.Popen(
        [
            "sudo",
            "powermetrics",
            "--samplers",
            "cpu_power",
            "-n",
            "1",
            "-i",
            "1",
            "-o",
            "/dev/null",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    _, stderr = process.communicate()

    if re.search(r"[sudo].*password", stderr):
        logger.debug(
            """Not using PowerMetrics, sudo password prompt detected.
                If you want to enable Powermetrics please modify your sudoers file
                as described in :
                https://mlco2.github.io/codecarbon/methodology.html#power-usage
            """
        )
        return False
    if process.returncode != 0:
        raise Exception("Return code != 0")
    else:
        return True


class ApplePowermetrics:
    _osx_silicon_exec = "powermetrics"

    def __init__(
        self,
        output_dir: str = ".",
        n_points=10,
        interval=100,
        log_file_name="powermetrics_log.txt",
    ):
        self._log_file_path = os.path.join(output_dir, log_file_name)
        self._system = sys.platform.lower()
        self._n_points = n_points
        self._interval = interval
        self._setup_cli()

    def _setup_cli(self):
        """
        Setup cli command to run Powermetrics
        """
        if self._system.startswith("darwin"):
            cpu_model = detect_cpu_model()
            if cpu_model.startswith("Apple"):
                if shutil.which(self._osx_silicon_exec):
                    self._cli = self._osx_silicon_exec
                else:
                    raise FileNotFoundError(
                        f"Powermetrics executable not found on {self._system}"
                    )
        else:
            raise SystemError("Platform not supported by Powermetrics")

    def _log_values(self):
        """
        Logs output from Powermetrics to a file
        """
        returncode = None

        if self._system.startswith("darwin"):
            # Run the powermetrics command with sudo and capture its output
            cmd = [
                "sudo",
                "powermetrics",
                "-n",
                str(self._n_points),
                "",
                "--samplers",
                "cpu_power",
                "--format",
                "csv",
                "-i",
                str(self._interval),
                "-o",
                self._log_file_path,
            ]
            returncode = subprocess.call(cmd, universal_newlines=True)

        else:
            return None

        if returncode != 0:
            logger.warning(
                "Returncode while logging power values using "
                + f"Powermetrics: {returncode}"
            )
        return

    def get_details(self, **kwargs) -> Dict:
        """
        Fetches the CPU Power Details by fetching values from a logged csv file
        in _log_values function
        """
        self._log_values()
        details = dict()
        try:
            with open(self._log_file_path) as f:
                logfile = f.read()
            cpu_pattern = r"CPU Power: (\d+) mW"
            cpu_power_list = re.findall(cpu_pattern, logfile)

            details["CPU Power"] = np.mean(
                [float(power) / 1000 for power in cpu_power_list]
            )
            details["CPU Energy Delta"] = np.sum(
                [
                    (self._interval / 1000) * (float(power) / 1000)
                    for power in cpu_power_list
                ]
            )
            gpu_pattern = r"GPU Power: (\d+) mW"
            gpu_power_list = re.findall(gpu_pattern, logfile)
            details["GPU Power"] = np.mean(
                [float(power) / 1000 for power in gpu_power_list]
            )
            details["GPU Energy Delta"] = np.sum(
                [
                    (self._interval / 1000) * (float(power) / 1000)
                    for power in gpu_power_list
                ]
            )
        except Exception as e:
            logger.info(
                f"Unable to read Powermetrics logged file at {self._log_file_path}\n \
                Exception occurred {e}",
                exc_info=True,
            )
        return details

    def start(self):
        # TODO: Read energy
        pass



from dataclasses import dataclass, field

import pynvml

from .units import Energy, Power, Time
from .logger import logger


@dataclass
class GPUDevice:
    handle: any
    gpu_index: int
    # Energy consumed in kWh
    energy_delta: Energy = field(default_factory=lambda: Energy(0))
    # Power based on reading
    power: Power = field(default_factory=lambda: Power(0))
    # Last energy reading in kWh
    last_energy: Energy = field(default_factory=lambda: Energy(0))

    def start(self):
        self.last_energy = self._get_energy_kwh()

    def __post_init__(self):
        self.last_energy = self._get_energy_kwh()
        self._init_static_details()

    def _get_energy_kwh(self):
        return Energy.from_millijoules(self._get_total_energy_consumption())

    def delta(self, duration: Time) -> dict:
        """
        Compute the energy/power used since last call.
        """
        new_last_energy = energy = self._get_energy_kwh()
        self.power = self.power.from_energies_and_delay(
            energy, self.last_energy, duration
        )
        self.energy_delta = energy - self.last_energy
        self.last_energy = new_last_energy
        return {
            "name": self._gpu_name,
            "uuid": self._uuid,
            "delta_energy_consumption": self.energy_delta,
            "power_usage": self.power,
        }

    def get_static_details(self):
        return {
            "name": self._gpu_name,
            "uuid": self._uuid,
            "total_memory": self._total_memory,
            "power_limit": self._power_limit,
            "gpu_index": self.gpu_index,
        }

    def _init_static_details(self):
        self._gpu_name = self._get_gpu_name()
        self._uuid = self._get_uuid()
        self._power_limit = self._get_power_limit()
        # Get the memory
        memory = self._get_memory_info()
        self._total_memory = memory.total

    def get_gpu_details(self):
        # Memory
        memory = self._get_memory_info()

        device_details = {
            "name": self._gpu_name,
            "uuid": self._uuid,
            "free_memory": memory.free,
            "total_memory": memory.total,
            "used_memory": memory.used,
            "temperature": self._get_temperature(),
            "power_usage": self._get_power_usage(),
            "power_limit": self._power_limit,
            "total_energy_consumption": self._get_total_energy_consumption(),
            "gpu_utilization": self._get_gpu_utilization(),
            "compute_mode": self._get_compute_mode(),
            "compute_processes": self._get_compute_processes(),
            "graphics_processes": self._get_graphics_processes(),
        }
        return device_details

    def _to_utf8(self, str_or_bytes):
        if hasattr(str_or_bytes, "decode"):
            return str_or_bytes.decode("utf-8", errors="replace")

        return str_or_bytes

    def _get_total_energy_consumption(self):
        """Returns total energy consumption for this GPU in millijoules (mJ) since the driver was last reloaded
        https://docs.nvidia.com/deploy/nvml-api/group__nvmlDeviceQueries.html#group__nvmlDeviceQueries_1g732ab899b5bd18ac4bfb93c02de4900a
        """
        return pynvml.nvmlDeviceGetTotalEnergyConsumption(self.handle)

    def _get_gpu_name(self):
        """Returns the name of the GPU device
        https://docs.nvidia.com/deploy/nvml-api/group__nvmlDeviceQueries.html#group__nvmlDeviceQueries_1ga5361803e044c6fdf3b08523fb6d1481
        """
        name = pynvml.nvmlDeviceGetName(self.handle)
        return self._to_utf8(name)

    def _get_uuid(self):
        """Returns the globally unique GPU device UUID
        https://docs.nvidia.com/deploy/nvml-api/group__nvmlDeviceQueries.html#group__nvmlDeviceQueries_1g72710fb20f30f0c2725ce31579832654
        """
        uuid = pynvml.nvmlDeviceGetUUID(self.handle)
        return self._to_utf8(uuid)

    def _get_memory_info(self):
        """Returns memory info in bytes
        https://docs.nvidia.com/deploy/nvml-api/group__nvmlDeviceQueries.html#group__nvmlDeviceQueries_1g2dfeb1db82aa1de91aa6edf941c85ca8
        """
        return pynvml.nvmlDeviceGetMemoryInfo(self.handle)

    def _get_temperature(self):
        """Returns degrees in the Celsius scale
        https://docs.nvidia.com/deploy/nvml-api/group__nvmlDeviceQueries.html#group__nvmlDeviceQueries_1g92d1c5182a14dd4be7090e3c1480b121
        """
        return pynvml.nvmlDeviceGetTemperature(self.handle, pynvml.NVML_TEMPERATURE_GPU)

    def _get_power_usage(self):
        """Returns power usage in milliwatts
        https://docs.nvidia.com/deploy/nvml-api/group__nvmlDeviceQueries.html#group__nvmlDeviceQueries_1g7ef7dff0ff14238d08a19ad7fb23fc87
        """
        return pynvml.nvmlDeviceGetPowerUsage(self.handle)

    def _get_power_limit(self):
        """Returns max power usage in milliwatts
        https://docs.nvidia.com/deploy/nvml-api/group__nvmlDeviceQueries.html#group__nvmlDeviceQueries_1g263b5bf552d5ec7fcd29a088264d10ad
        """
        try:
            return pynvml.nvmlDeviceGetEnforcedPowerLimit(self.handle)
        except Exception:
            return None

    def _get_gpu_utilization(self):
        """Returns the % of utilization of the kernels during the last sample
        https://docs.nvidia.com/deploy/nvml-api/structnvmlUtilization__t.html#structnvmlUtilization__t
        """
        return pynvml.nvmlDeviceGetUtilizationRates(self.handle).gpu

    def _get_compute_mode(self):
        """Returns the compute mode of the GPU
        https://docs.nvidia.com/deploy/nvml-api/group__nvmlDeviceEnumvs.html#group__nvmlDeviceEnumvs_1gbed1b88f2e3ba39070d31d1db4340233
        """
        return pynvml.nvmlDeviceGetComputeMode(self.handle)

    def _get_compute_processes(self):
        """Returns the list of processes ids having a compute context on the
        device with the memory used
        https://docs.nvidia.com/deploy/nvml-api/group__nvmlDeviceQueries.html#group__nvmlDeviceQueries_1g46ceaea624d5c96e098e03c453419d68
        """
        try:
            processes = pynvml.nvmlDeviceGetComputeRunningProcesses(self.handle)

            return [{"pid": p.pid, "used_memory": p.usedGpuMemory} for p in processes]
        except pynvml.NVMLError:
            return []

    def _get_graphics_processes(self):
        """Returns the list of processes ids having a graphics context on the
        device with the memory used
        https://docs.nvidia.com/deploy/nvml-api/group__nvmlDeviceQueries.html#group__nvmlDeviceQueries_1g7eacf7fa7ba4f4485d166736bf31195e
        """
        try:
            processes = pynvml.nvmlDeviceGetGraphicsRunningProcesses(self.handle)

            return [{"pid": p.pid, "used_memory": p.usedGpuMemory} for p in processes]
        except pynvml.NVMLError:
            return []


class AllGPUDevices:
    def __init__(self):
        if is_gpu_details_available():
            logger.debug("GPU available. Starting setup")
            self.device_count = pynvml.nvmlDeviceGetCount()
        else:
            logger.error("There is no GPU available")
            self.device_count = 0
        self.devices = []
        for i in range(self.device_count):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            gpu_device = GPUDevice(handle=handle, gpu_index=i)
            self.devices.append(gpu_device)

    def get_gpu_static_info(self):
        """Get all GPUs static information.
        >>> get_gpu_static_info()
        [
            {
                "name": "Tesla V100-SXM2-16GB",
                "uuid": "GPU-4e817856-1fb8-192a-7ab7-0e0e4476c184",
                "total_memory": 16945512448,
                "power_limit": 300000,
                "gpu_index": 0,
            }
        ]
        """
        try:
            devices_static_info = []
            for i in range(self.device_count):
                gpu_device = self.devices[i]
                devices_static_info.append(gpu_device.get_static_details())
            return devices_static_info

        except pynvml.NVMLError:
            logger.warning("Failed to retrieve gpu static info", exc_info=True)
            return []

    def get_gpu_details(self):
        """Get all GPUs instantaneous metrics
        >>> get_gpu_details()
        [
            {
                "name": "Tesla V100-SXM2-16GB",
                "uuid": "GPU-4e817856-1fb8-192a-7ab7-0e0e4476c184",
                "free_memory": 16945381376,
                "total_memory": 16945512448,
                "used_memory": 131072,
                "temperature": 28,
                "total_energy_consumption":2000,
                "power_usage": 42159,
                "power_limit": 300000,
                "gpu_utilization": 0,
                "compute_mode": 0,
                "compute_processes": [],
                "graphics_processes": [],
            }
        ]
        """
        try:
            devices_info = []
            for i in range(self.device_count):
                gpu_device: GPUDevice = self.devices[i]
                devices_info.append(gpu_device.get_gpu_details())
            return devices_info

        except pynvml.NVMLError:
            logger.warning("Failed to retrieve gpu information", exc_info=True)
            return []

    def get_delta(self, last_duration: Time):
        """Get difference since last time this function was called
        >>> get_delta()
        [
            {
                "name": "Tesla V100-SXM2-16GB",
                "uuid": "GPU-4e817856-1fb8-192a-7ab7-0e0e4476c184",
                "delta_energy_consumption":2000,
                "power_usage": 42159,
            }
        ]
        """
        try:
            devices_info = []
            for i in range(self.device_count):
                gpu_device: GPUDevice = self.devices[i]
                devices_info.append(gpu_device.delta(last_duration))
            return devices_info

        except pynvml.NVMLError:
            logger.warning("Failed to retrieve gpu information", exc_info=True)
            return []


def is_gpu_details_available():
    """Returns True if the GPU details are available."""
    try:
        pynvml.nvmlInit()
        return True

    except pynvml.NVMLError:
        return False



import time
from uuid import uuid4

from codecarbon.output import EmissionsData, TaskEmissionsData


class Task:
    """
    A task, used to segregate electrical consumption when executing a treatment.
    """

    is_active: bool
    emissions_data: EmissionsData

    def __init__(self, task_name):  # , task_measure
        self.task_id: str = task_name + uuid4().__str__()
        self.task_name: str = task_name
        self.start_time = time.time()
        self.is_active = True

    def out(self):
        return TaskEmissionsData(
            task_name=self.task_name,
            timestamp=self.emissions_data.timestamp,
            project_name=self.emissions_data.project_name,
            run_id=self.emissions_data.run_id,
            duration=self.emissions_data.duration,
            emissions=self.emissions_data.emissions,
            emissions_rate=self.emissions_data.emissions_rate,
            cpu_power=self.emissions_data.cpu_power,
            gpu_power=self.emissions_data.gpu_power,
            ram_power=self.emissions_data.ram_power,
            cpu_energy=self.emissions_data.cpu_energy,
            gpu_energy=self.emissions_data.gpu_energy,
            ram_energy=self.emissions_data.ram_energy,
            energy_consumed=self.emissions_data.energy_consumed,
            country_name=self.emissions_data.country_name,
            country_iso_code=self.emissions_data.country_iso_code,
            region=self.emissions_data.region,
            cloud_provider=self.emissions_data.cloud_provider,
            cloud_region=self.emissions_data.cloud_region,
            os=self.emissions_data.os,
            python_version=self.emissions_data.python_version,
            codecarbon_version=self.emissions_data.codecarbon_version,
            cpu_count=self.emissions_data.cpu_count,
            cpu_model=self.emissions_data.cpu_model,
            gpu_count=self.emissions_data.gpu_count,
            gpu_model=self.emissions_data.gpu_model,
            longitude=self.emissions_data.longitude,
            latitude=self.emissions_data.latitude,
            ram_total_size=self.emissions_data.ram_total_size,
            tracking_mode=self.emissions_data.tracking_mode,
            on_cloud=self.emissions_data.on_cloud,
        )



from flask import Flask 
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.orm import Mapped, mapped_column
#import geocoder

### APP CONFIGURATION ###
app = Flask(__name__)
app.config['EXPLAIN_TEMPLATE_LOADING'] = True
app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///sit.db' 
db = SQLAlchemy(app)

### DATABASE MODELS ###
train_devices = db.Table('train_devices',
    db.Column('hostgpu_id', db.Integer, db.ForeignKey('hostgpu.id'), primary_key=True),
    db.Column('tracking_info_id', db.String, db.ForeignKey('trackinginfo.id'), primary_key=True),
    db.Column('CO2_usage', db.Float, nullable=True),
    db.Column('power_usg', db.Float, nullable=True)
)

inference_devices = db.Table('inference_devices',
    db.Column('hostgpu_id', db.Integer, db.ForeignKey('hostgpu.id'), primary_key=True),
    db.Column('inference_info_id', db.String, db.ForeignKey('inferenceinfo.id'), primary_key=True),
    db.Column('CO2_usage', db.Float, nullable=True)
)

class Trackinginfo(db.Model):
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    timestamp: Mapped[int] = mapped_column(unique=True)
    carbon_intensity: Mapped[float] = mapped_column(unique=False)
    power_usage: Mapped[float] = mapped_column(unique=False)
    CO2_usage: Mapped[float] = mapped_column(unique=False)
    model: Mapped[str] = mapped_column(unique=False)
    elapsed_time: Mapped[float] = mapped_column(unique=False)
    CO2_inference: Mapped[float] = mapped_column(unique=False, nullable=True)
    task_model: Mapped[str] = mapped_column(unique=False)
    task_name = db.Column(db.Integer, db.ForeignKey('task.task'), nullable=False)
    train_devices = db.relationship('Hostgpu', secondary=train_devices, lazy='subquery',
            backref=db.backref('trackinginfo', lazy=True))

    def __repr__(self):
        return f"Trackinginfo(id={self.id}, timestamp={self.timestamp}, " \
               f"carbon_intensity={self.carbon_intensity}, power_usage={self.power_usage}, " \
               f"CO2_usage={self.CO2_usage}, task_name={self.task_name}, model={self.model}, " \
               f"elapsed_time={self.elapsed_time}, CO2_inference={self.CO2_inference}, task_model={self.task_model})"
    
class Inferenceinfo(db.Model):
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    timestamp_inf: Mapped[int] = mapped_column(unique=True)
    carbon_intensity_inf: Mapped[float] = mapped_column(unique=False)
    power_usage_inf: Mapped[float] = mapped_column(unique=False)
    CO2_usage_inf: Mapped[float] = mapped_column(unique=False)
    model_inf: Mapped[str] = mapped_column(unique=False)
    task_model_inf: Mapped[str] = mapped_column(unique=False)
    elapsed_time_inf: Mapped[float] = mapped_column(unique=False)
    task_name_inf = db.Column(db.Integer, db.ForeignKey('task.task'), nullable=False)
    inf_devices = db.relationship('Hostgpu', secondary=inference_devices, lazy='subquery',
        backref=db.backref('inferenceinfo', lazy=True))

    def __repr__(self):
        return f"Inferenceinfo(id={self.id}, timestamp_inf={self.timestamp_inf}, " \
                f"carbon_intensity_inf={self.carbon_intensity_inf}, power_usage_inf={self.power_usage_inf}, " \
                f"CO2_usage_inf={self.CO2_usage_inf}, task_name_inf={self.task_name_inf}, " \
                f"model_inf={self.model_inf}, elapsed_time_inf={self.elapsed_time_inf}, task_model_inf={self.task_model_inf})"

hostgpus = db.Table('hostgpus',
    db.Column('hostgpu_host', db.Integer, db.ForeignKey('hostgpu.id'), primary_key=True),
    db.Column('task_id', db.String, db.ForeignKey('task.task'), primary_key=True)
)

class Task(db.Model):
    #id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    task: Mapped[str] = mapped_column(primary_key=True, unique=True)
    track_infos = db.relationship('Trackinginfo', backref='tasks', lazy=True)
    inference_infos = db.relationship('Inferenceinfo', backref='tasks', lazy=True)

    tags = db.relationship('Hostgpu', secondary=hostgpus, lazy='subquery',
        backref=db.backref('tasks', lazy=True))
    
    models = db.relationship('Model', backref='tasks', lazy=True)

    def __repr__(self):
        return f"Task(task={self.task})"

class Model(db.Model):
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    model_name: Mapped[str] = mapped_column(unique=False)
    task = db.Column(db.Integer, db.ForeignKey('task'), nullable=False)

    def __repr__(self):
        return f"Model(id={self.id}, model_name={self.model_name}, task_id={self.task})"

class Hostgpu(db.Model):
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    host: Mapped[str] = mapped_column(unique=False)
    gpu_cpu: Mapped[str] = mapped_column(unique=False)

    def __repr__(self):
        return f"HostGPU(id={self.id}, host={self.host}, gpu_model={self.gpu_cpu})"


class Powermix(db.Model):
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    timestamp: Mapped[int] = mapped_column(unique=False)
    powermix: Mapped[str] = mapped_column(unique=False)

    def __repr__(self):
        return f"Powermix(id={self.id}, timestamp={self.timestamp}, powermix={self.powermix})"


from dataclasses import dataclass, field

from .units import Energy, Power, Time
from .logger import logger


@dataclass
class RAPLFile:
    # RAPL device being measured
    name: str
    # Path to file containing RAPL reading
    path: str
    # Path to corresponding file containing maximum possible RAPL reading
    max_path: str
    # Energy consumed in kWh
    energy_delta: Energy = field(default_factory=lambda: Energy(0))
    # Power based on reading
    power: Power = field(default_factory=lambda: Power(0))
    # Last energy reading in kWh
    last_energy: Energy = field(default_factory=lambda: Energy(0))
    # Max value energy can hold before it wraps
    max_energy_reading: Energy = field(default_factory=lambda: Energy(0))

    def __post_init__(self):
        self.last_energy = self._get_value()
        with open(self.max_path, "r") as f:
            max_micro_joules = float(f.read())

            self.max_energy_reading = Energy.from_ujoules(max_micro_joules)

    def _get_value(self) -> Energy:
        """
        Reads the value in the file at the path
        """
        with open(self.path, "r") as f:
            micro_joules = float(f.read())

            e = Energy.from_ujoules(micro_joules)
            return e

    def start(self) -> None:
        self.last_energy = self._get_value()
        return

    def delta(self, duration: Time) -> None:
        """
        Compute the energy used since last call.
        """
        new_last_energy = energy = self._get_value()
        if self.last_energy > energy:
            logger.debug(
                f"In RAPLFile : Current energy value ({energy}) is lower than previous value ({self.last_energy}). Assuming wrap-around! Source file : {self.path}"
            )
            energy = energy + self.max_energy_reading
        self.power = self.power.from_energies_and_delay(
            energy, self.last_energy, duration
        )
        self.energy_delta = energy - self.last_energy
        self.last_energy = new_last_energy

        return



from setuptools import find_packages, setup

with open("app/README.md", "r") as f:
    long_description = f.read()

setup(
    name="carbontracking",
    version="0.0.31",
    description="A carbon tracking library for Python applications",
    package_dir={"": "app"},
    packages=find_packages(where="app"),
    package_data={'': ['templates/index.html', 'static/assets/bootstrap/css/*', 'static/assets/bootstrap/js/*', 'static/assets/fonts/*', 'static/assets/img/*', 'static/assets/js/*']},
    include_package_data=True,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    author="Peter Rolfes",
    author_email="pero01@dfki.de",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'codecarbon',
        'requests',
        'numpy',
        'pandas',
        'psutil',
        'rapidfuzz',
        'flask',
        'flask_sqlalchemy',
        'sqlalchemy',
        'statistics',
        'collection',
        'thread',
        'jsons',
        'codecarbon',
        'entsoe-py',
    ],
    extras_require={
        "dev": ["pytest>=7.0", "twine>=4.0.2"],
    },
    python_requires=">=3.10",
)


"""
Implements tracking Intel CPU Power Consumption on Mac and Windows
using Intel Power Gadget
https://software.intel.com/content/www/us/en/develop/articles/intel-power-gadget.html
"""
import os
import shutil
import subprocess
import sys
from typing import Dict, Tuple

import pandas as pd
from rapidfuzz import fuzz, process, utils

from .rapl import RAPLFile
from .units import Time
from .util import detect_cpu_model
from .logger import logger
from .input import DataSource


def is_powergadget_available():
    try:
        IntelPowerGadget()
        return True
    except Exception as e:
        logger.debug(
            "Not using PowerGadget, an exception occurred while instantiating"
            + f" IntelPowerGadget : {e}",
        )
        return False


def is_rapl_available():
    try:
        IntelRAPL()
        return True
    except Exception as e:
        logger.debug(
            "Not using the RAPL interface, an exception occurred while instantiating "
            + f"IntelRAPL : {e}",
        )
        return False


class IntelPowerGadget:
    _osx_exec = "PowerLog"
    _osx_exec_backup = "/Applications/Intel Power Gadget/PowerLog"
    _windows_exec = "PowerLog3.0.exe"

    def __init__(
        self,
        output_dir: str = ".",
        duration=1,
        resolution=100,
        log_file_name="intel_power_gadget_log.csv",
    ):
        self._log_file_path = os.path.join(output_dir, log_file_name)
        self._system = sys.platform.lower()
        self._duration = duration
        self._resolution = resolution
        self._setup_cli()

    def _setup_cli(self):
        """
        Setup cli command to run Intel Power Gadget
        """
        if self._system.startswith("win"):
            self._get_windows_exec_backup()
            if shutil.which(self._windows_exec):
                self._cli = shutil.which(
                    self._windows_exec
                )  # Windows exec is a relative path
            elif shutil.which(self._windows_exec_backup):
                self._cli = self._windows_exec_backup
            else:
                raise FileNotFoundError(
                    f"Intel Power Gadget executable not found on {self._system}"
                )
        elif self._system.startswith("darwin"):
            if shutil.which(self._osx_exec):
                self._cli = self._osx_exec
            elif shutil.which(self._osx_exec_backup):
                self._cli = self._osx_exec_backup
            else:
                raise FileNotFoundError(
                    f"Intel Power Gadget executable not found on {self._system}"
                )
        else:
            raise SystemError("Platform not supported by Intel Power Gadget")

    def _get_windows_exec_backup(self) -> None:
        """
        Find the windows executable for the current version of intel power gadget.
        Example: "C:\\Program Files\\Intel\\Power Gadget 3.5\\PowerLog3.0.exe"
        """
        parent_folder = "C:\\Program Files\\Intel\\"

        # Get a list of all subdirectories in the parent folder
        subfolders = [f.name for f in os.scandir(parent_folder) if f.is_dir()]

        # Look for a folder that contains "Power Gadget" in its name
        desired_folder = next(
            (folder for folder in subfolders if "Power Gadget" in folder), None
        )
        if desired_folder:
            self._windows_exec_backup = os.path.join(
                parent_folder, desired_folder, self._windows_exec
            )
        else:
            self._windows_exec_backup = None

    def _log_values(self):
        """
        Logs output from Intel Power Gadget command line to a file
        """
        returncode = None
        if self._system.startswith("win"):
            returncode = subprocess.call(
                [
                    self._cli,
                    "-duration",
                    str(self._duration),
                    "-resolution",
                    str(self._resolution),
                    "-file",
                    self._log_file_path,
                ],
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        elif self._system.startswith("darwin"):
            returncode = subprocess.call(
                f"'{self._cli}' -duration {self._duration} -resolution {self._resolution} -file {self._log_file_path} > /dev/null",  # noqa: E501
                shell=True,
            )
        else:
            return None

        if returncode != 0:
            logger.warning(
                "Returncode while logging power values using "
                + f"Intel Power Gadget: {returncode}"
            )
        return

    def get_cpu_details(self, **kwargs) -> Dict:
        """
        Fetches the CPU Power Details by fetching values from a logged csv file
        in _log_values function
        """
        self._log_values()
        cpu_details = dict()
        try:
            cpu_data = pd.read_csv(self._log_file_path).dropna()
            for col_name in cpu_data.columns:
                if col_name in ["System Time", "Elapsed Time (sec)", "RDTSC"]:
                    continue
                if "Cumulative" in col_name:
                    cpu_details[col_name] = cpu_data[col_name].iloc[-1]
                else:
                    cpu_details[col_name] = cpu_data[col_name].mean()
        except Exception as e:
            logger.info(
                f"Unable to read Intel Power Gadget logged file at {self._log_file_path}\n \
                Exception occurred {e}",
                exc_info=True,
            )
        return cpu_details

    def start(self):
        # TODO: Read energy
        pass


class IntelRAPL:
    def __init__(self, rapl_dir="/sys/class/powercap/intel-rapl"):
        self._lin_rapl_dir = rapl_dir
        self._system = sys.platform.lower()
        self._rapl_files = list()
        self._setup_rapl()
        self._cpu_details: Dict = dict()

        self._last_mesure = 0

    def _is_platform_supported(self) -> bool:
        return self._system.startswith("lin")

    def _setup_rapl(self):
        if self._is_platform_supported():
            if os.path.exists(self._lin_rapl_dir):
                self._fetch_rapl_files()
            else:
                raise FileNotFoundError(
                    f"Intel RAPL files not found at {self._lin_rapl_dir} "
                    + f"on {self._system}"
                )
        else:
            raise SystemError("Platform not supported by Intel RAPL Interface")
        return

    def _fetch_rapl_files(self):
        """
        Fetches RAPL files from the RAPL directory
        """

        # consider files like `intel-rapl:$i`
        files = list(filter(lambda x: ":" in x, os.listdir(self._lin_rapl_dir)))

        i = 0
        for file in files:
            path = os.path.join(self._lin_rapl_dir, file, "name")
            with open(path) as f:
                name = f.read().strip()
                # Fake the name used by Power Gadget
                if "package" in name:
                    name = f"Processor Energy Delta_{i}(kWh)"
                    i += 1
                # RAPL file to take measurement from
                rapl_file = os.path.join(self._lin_rapl_dir, file, "energy_uj")
                # RAPL file containing maximum possible value of energy_uj above which it wraps
                rapl_file_max = os.path.join(
                    self._lin_rapl_dir, file, "max_energy_range_uj"
                )
                try:
                    # Try to read the file to be sure we can
                    with open(rapl_file, "r") as f:
                        _ = float(f.read())
                    self._rapl_files.append(
                        RAPLFile(name=name, path=rapl_file, max_path=rapl_file_max)
                    )
                    logger.debug(f"We will read Intel RAPL files at {rapl_file}")
                except PermissionError as e:
                    raise PermissionError(
                        "Unable to read Intel RAPL files for CPU power, we will use a constant for your CPU power."
                        + " Please view https://github.com/mlco2/codecarbon/issues/244"
                        + f" for workarounds : {e}"
                    )
        return

    def get_cpu_details(self, duration: Time, **kwargs) -> Dict:
        """
        Fetches the CPU Energy Deltas by fetching values from RAPL files
        """
        cpu_details = dict()
        try:
            list(map(lambda rapl_file: rapl_file.delta(duration), self._rapl_files))

            for rapl_file in self._rapl_files:
                logger.debug(rapl_file)
                cpu_details[rapl_file.name] = rapl_file.energy_delta.kWh
                # We fake the name used by Power Gadget when using RAPL
                if "Energy" in rapl_file.name:
                    cpu_details[
                        rapl_file.name.replace("Energy", "Power")
                    ] = rapl_file.power.W
        except Exception as e:
            logger.info(
                f"Unable to read Intel RAPL files at {self._rapl_files}\n \
                Exception occurred {e}",
                exc_info=True,
            )
        self.cpu_details = cpu_details
        logger.debug(f"get_cpu_details {self.cpu_details}")
        return cpu_details

    def get_static_cpu_details(self) -> Dict:
        """
        Return CPU details without computing them.
        """
        logger.debug(f"get_static_cpu_details {self.cpu_details}")

        return self.cpu_details

    def start(self):
        for rapl_file in self._rapl_files:
            rapl_file.start()


class TDP:
    def __init__(self):
        self.model, self.tdp = self._main()

    @staticmethod
    def _get_cpu_constant_power(match: str, cpu_power_df: pd.DataFrame) -> int:
        """Extract constant power from matched CPU"""
        return cpu_power_df[cpu_power_df["Name"] == match]["TDP"].values[0]

    def _get_cpu_power_from_registry(self, cpu_model_raw: str) -> int:
        cpu_power_df = DataSource().get_cpu_power_data()
        cpu_matching = self._get_matching_cpu(cpu_model_raw, cpu_power_df)
        if cpu_matching:
            power = self._get_cpu_constant_power(cpu_matching, cpu_power_df)
            return power
        return None

    def _get_matching_cpu(
        self, model_raw: str, cpu_df: pd.DataFrame, greedy=False
    ) -> str:
        """
        Get matching cpu name

        :args:
            model_raw (str): raw name of the cpu model detected on the machine

            cpu_df (DataFrame): table containing cpu models along their tdp

            greedy (default False): if multiple cpu models match with an equal
            ratio of similarity, greedy (True) selects the first model,
            following the order of the cpu list provided, while non-greedy
            returns None.

        :return: name of the matching cpu model

        :notes:
            Thanks to the greedy mode, even though the match could be a model
            with a tdp very different from the actual tdp of current cpu, it
            still enables the relative comparison of models emissions running
            on the same machine.

            THRESHOLD_DIRECT defines the similarity ratio value to consider
            almost-exact matches.

            THRESHOLD_TOKEN_SET defines the similarity ratio value to consider
            token_set matches (for more detail see fuzz.token_set_ratio).
        """
        THRESHOLD_DIRECT = 100
        THRESHOLD_TOKEN_SET = 100

        direct_match = process.extractOne(
            model_raw,
            cpu_df["Name"],
            processor=lambda s: s.lower(),
            scorer=fuzz.ratio,
            score_cutoff=THRESHOLD_DIRECT,
        )

        if direct_match:
            return direct_match[0]

        indirect_matches = process.extract(
            model_raw,
            cpu_df["Name"],
            processor=utils.default_process,
            scorer=fuzz.token_set_ratio,
            score_cutoff=THRESHOLD_TOKEN_SET,
        )

        if indirect_matches:
            if (
                greedy
                or len(indirect_matches) == 1
                or indirect_matches[0][1] != indirect_matches[1][1]
            ):
                return indirect_matches[0][0]

        return None

    def _main(self) -> Tuple[str, int]:
        """
        Get CPU power from constant mode

        :return: model name (str), power in Watt (int)
        """
        cpu_model_detected = detect_cpu_model()

        if cpu_model_detected:
            power = self._get_cpu_power_from_registry(cpu_model_detected)

            if power:
                logger.debug(
                    f"CPU : We detect a {cpu_model_detected} with a TDP of {power} W"
                )
                return cpu_model_detected, power
            logger.warning(
                f"We saw that you have a {cpu_model_detected} but we don't know it."
                + " Please contact us."
            )
            return cpu_model_detected, None
        logger.warning(
            "We were unable to detect your CPU using the `cpuinfo` package."
            + " Resorting to a default power consumption of 85W."
        )
        return "Unknown", None

    def start(self):
        pass
